import path from "node:path";
import { Glob } from "bun";
import YAML from "yaml";
import z from "zod";

const BindingSchema = z.object({
  $binding: z.literal("playbill"),
  id: z.string(),
  name: z.string().default("Unnamed playbill"),
  author: z.string().optional(),
  version: z.string(),
  files: z.array(z.string()).default(["**/*.yaml"]),
});

type Binding = z.infer<typeof BindingSchema>;

export interface PlaybillBinding {
  info: Binding;
  bindingPath: string;
  files: string[];
}

export async function loadBindingFile(
  filePath: string,
): Promise<PlaybillBinding> {
  const file = Bun.file(filePath);
  const text = await file.text();
  const parsed = YAML.parse(text);
  const binding = BindingSchema.parse(parsed);

  const fileGlobs = binding.files;
  const bindingFileDirname = filePath.split("/").slice(0, -1).join("/");

  const allFilePaths: string[] = [];

  for (const thisGlob of fileGlobs) {
    const glob = new Glob(thisGlob);

    const results = glob.scanSync({
      cwd: bindingFileDirname,
      absolute: true,
      followSymlinks: true,
      onlyFiles: true,
    });

    allFilePaths.push(...results);
  }

  const bindingFileAbsolutePath = path.resolve(filePath);
  const filteredFilePaths = allFilePaths.filter((filePath) => {
    return filePath !== bindingFileAbsolutePath;
  });

  return {
    info: binding,
    bindingPath: filePath,
    files: filteredFilePaths,
  };
}
