import { z } from "zod";
import { StatsSchema, type Playbill } from "./playbill-schema";

const DirectiveSchema = z.enum(["dmg", "ability", "stat", "rule"] as const);
type DirectiveName = z.infer<typeof DirectiveSchema>;
type DirectiveHandler = (playbill: Playbill, ...args: string[]) => string;

export class DirectiveError extends Error { }

const ProsceniumLangDirectives: Record<DirectiveName, DirectiveHandler> = {
  dmg: (_, amount = "1", type = "phy") => {
    const DAMAGE_TYPES = ["phy", "arc", "physical", "arcane"];
    if (!DAMAGE_TYPES.includes(type)) {
      throw new DirectiveError(`Unknown damage type: ${type}`);
    }

    const finalizedType = type.slice(0, 3);
    return `<em class='dmg ${finalizedType}'>${amount}</em>`;
  },

  ability: (playbill: Playbill, ref: string) => {
    if (typeof ref !== "string") {
      throw new DirectiveError("Invalid ability reference");
    }

    const foundAbility = playbill.abilities.find((a) => a.id === ref);

    if (foundAbility === undefined) {
      throw new DirectiveError(`Unknown ability reference: ${ref}`);
    }

    return `<em class="ref-ability">${foundAbility.name}</em>`;
  },

  rule: (playbill: Playbill, ref: string) => {
    if (typeof ref !== "string") {
      throw new DirectiveError("Invalid rule reference");
    }

    const foundRule = playbill.rules.find((r) => r.id === ref);

    if (foundRule === undefined) {
      throw new DirectiveError(`Unknown ability reference: ${ref}`);
    }

    return `<em class="ref-rule">${foundRule.name}</em>`;
  },
  stat: (_, stat: string, amount = "") => {
    const lower = stat.toLowerCase();
    const parsed = StatsSchema.safeParse(lower);

    if (!parsed.success) {
      throw new DirectiveError(`Unknown stat: ${stat}`);
    }

    if (amount !== "") {
      const amountAsNum = Number.parseInt(amount, 10);
      if (Number.isNaN(amountAsNum)) {
        throw new DirectiveError(`Invalid amount: ${amount}`);
      }
      return `<em class="stat">${amount} ${stat.toUpperCase()}</em>`;
    }

    return `<em class="stat">${stat.toUpperCase()}</em>`;
  },
};

function processCommands(text: string, playbill: Playbill): string {
  const commandPattern = /\{(\w+):([^}]+)\}/g;

  return text.replace(commandPattern, (_match, command, args) => {
    const directive = DirectiveSchema.parse(command);
    const handler = ProsceniumLangDirectives[directive];
    const parsedArgs = [];
    let currentArg = "";
    let inQuotes = false;

    for (let i = 0; i < args.length; i++) {
      const char = args[i];
      if (char === '"' && (i === 0 || args[i - 1] !== "\\")) {
        inQuotes = !inQuotes;
      } else if (char === "," && !inQuotes) {
        parsedArgs.push(currentArg.trim().replace(/^"|"$/g, ""));
        currentArg = "";
      } else {
        currentArg += char;
      }
    }
    parsedArgs.push(currentArg.trim().replace(/^"|"$/g, ""));

    return handler(playbill, ...parsedArgs);
  });
}

export async function processLang(
  input: string,
  playbill: Playbill,
): Promise<string> {
  const lines = input.split("\n");
  const processedLines: string[] = [];
  for (const line of lines) {
    const processed = processCommands(line, playbill);
    processedLines.push(processed);
  }

  return processedLines.join("\n");
}
