import YAML from "yaml";
import {
  type AnyResource,
  AnyResourceSchema,
  ResourceMap,
} from "./playbill-schema";

export class ResourceFileError extends Error { }
export class NullResourceError extends Error { }

/**
 * Load and process all documents in a yaml file at the given path
 * @param filePath - The path to the yaml file
 * @returns An array of validated resources
 *
 * @throws ResourceFileError
 * @throws NullResourceError
 */
export async function loadResourceFile(
  filePath: string,
): Promise<AnyResource[]> {
  const file = Bun.file(filePath);
  const text = await file.text();

  const parsedDocs = YAML.parseAllDocuments(text);

  if (parsedDocs.some((doc) => doc.toJS() === null)) {
    throw new NullResourceError(`Null resource defined in ${filePath}`);
  }

  const errors = parsedDocs.flatMap((doc) => doc.errors);

  if (errors.length > 0) {
    throw new ResourceFileError(`Error parsing ${filePath}: ${errors}`);
  }

  const collection: AnyResource[] = [];

  for (const doc of parsedDocs) {
    if (doc.errors.length > 0) {
      throw new ResourceFileError(`Error parsing ${filePath}: ${doc.errors}`);
    }

    const raw = doc.toJS();
    const parsed = AnyResourceSchema.parse(raw);
    const type = parsed.$define;
    const schemaToUse = ResourceMap[type];
    const validated = schemaToUse.parse(parsed);
    collection.push(validated);
  }

  return collection;
}
